package com.coravy.wiimote {
    
	import flash.errors.IllegalOperationError;
	import flash.events.Event;
	import flash.events.EventDispatcher;
	import flash.events.IEventDispatcher;
	import flash.events.TimerEvent;
	import flash.utils.Timer;
	
	import org.wiiflash.Wiimote;
	import org.wiiflash.events.WiimoteEvent;
    
	/** 
	 * @eventType com.coravy.wiimote.GestureEvent.JERK_UP
 	 */	
	[Event(name=GestureEvent.JERK_UP, type='com.coravy.wiimote.GestureEvent')]
    
	/** 
	 * @eventType com.coravy.wiimote.GestureEvent.JERK_DOWN
 	 */	
	[Event(name=GestureEvent.JERK_DOWN, type='com.coravy.wiimote.GestureEvent')]
	
	/** 
	 * @eventType com.coravy.wiimote.GestureEvent.JERK_LEFT
 	 */	
	[Event(name=GestureEvent.JERK_LEFT, type='com.coravy.wiimote.GestureEvent')]
	
	/** 
	 * @eventType com.coravy.wiimote.GestureEvent.JERK_RIGHT
 	 */	
	[Event(name=GestureEvent.JERK_RIGHT, type='com.coravy.wiimote.GestureEvent')]
    
    /**
     * Einfache Klasse zur Gestenerkennung mit der Wiimote.
     * <p>
     * Gesten sind applikationsspezifisch. Diese Klasse erkennt nur die folgenden (einfachen)
     * Gesten:
     * <ul>
     *    <li>Schnelle Bewegung (Schlag) nach rechts   JERK_RIGHT.</li>
     *    <li>Schnelle Bewegung (Schlag) nach links    JERK_LEFT.</li>
     *    <li>Schnelle Bewegung (Schlag) nach oben     JERK_UP.</li>
     *    <li>Schnelle Bewegung (Schlag) nach unten    JERK_DOWN.</li>
     * </ul>
     * 
     * Für die Gestenerkennung werden ausschliesslich die Daten des Accelerometers benutzt.
     * 
     * @author Stefan Saasen <stefan@coravy.com> 
     */    
    public final class GestureDetection implements IEventDispatcher {
        
        /**
        * Filterfaktor für den Hochpassfilter.
        */
        private static const FILTERING_FACTOR:Number = 0.1;
        
        /**
        * Schwellwert: Bei Überschreitung wird die Wiimote als in Bewegung befindlich erkannt.
        */
        private static const REST_THRESHOLD:Number = 1.2;
        
        /**
        * Hilfskonstanten für die Gesture detection.
        */
        private static const MOTION_DETECTION_TIMEOUT:int = 450; // milliseconds
        private static const MOTION_THRESHOLD:Number = 1.55;
        private var  _motionUp:Boolean = false;
        private var  _motionDown:Boolean = false;
        private var  _motionLeft:Boolean = false;
        private var  _motionRight:Boolean = false;
        
        /**
        * Der EventDispatcher dieser Klasse.
        */
        private var _eventDispatcher:EventDispatcher;
        
        /**
        * Die Wiimote. Die Verbindung wird bereits ausserhalb dieser Klasse aufgebaut.
        */
        private var _wiimote:Wiimote;
        
        [ArrayElementType('Number')]
        private var _accelerationVec:Array = [0.0, 0.0, 0.0];
        
        private var motionTimer:Timer;
        
        public function GestureDetection(wiimote:Wiimote) {
            this._wiimote = wiimote;
            _wiimote.addEventListener(WiimoteEvent.UPDATE, onMotionUpdate); 
            _eventDispatcher = new EventDispatcher();
            
            motionTimer = new Timer(MOTION_DETECTION_TIMEOUT, 1);
            motionTimer.addEventListener(TimerEvent.TIMER, function():void {
                _motionUp = _motionDown = _motionRight = _motionLeft = false;
                motionTimer.reset();
            });
        }
               
        /**
        * Verarbeitet die eingehenden Rohdaten der Wiimote.
        */
        private function onMotionUpdate(e:WiimoteEvent):void {
            // Die ungefilterten Eingangsdaten der Wiimote.
		    var x:Number  = e.target.sensorX;
            var y:Number  = e.target.sensorY;
		    var z:Number  = e.target.sensorZ;
		    
            // Berechnung der euklidischen Distanz - In Ruhelage ca. 1g
		    var length:Number = Math.sqrt(x * x + y * y + z * z);
		    
            // Hoch-Pass-Filter
            _accelerationVec[0] = x - (_accelerationVec[0] = x * FILTERING_FACTOR + _accelerationVec[0] * (1.0 - FILTERING_FACTOR));
            _accelerationVec[1] = y - (_accelerationVec[1] = y * FILTERING_FACTOR + _accelerationVec[1] * (1.0 - FILTERING_FACTOR));
            _accelerationVec[2] = z - (_accelerationVec[2] = z * FILTERING_FACTOR + _accelerationVec[2] * (1.0 - FILTERING_FACTOR));
            
            // Wir sind nur an Bewegungsdaten interessiert. Wenn sich die Wiimote in Ruhe befindet (leichte Bewegungen sind
            // dabei eingeschlossen) werden die Daten gar nicht erst weitergegeben.
            if(length > REST_THRESHOLD) {
                afterFilter(_accelerationVec[0], _accelerationVec[1], _accelerationVec[2]);  
            }
        }
        
       /**
        * Verarbeitet die gefilterten Daten der Wiimote
        */
        private function afterFilter(x:Number, y:Number, z:Number):void {
    		// Bei einer ruckartigen Bewegung kann eine Zunahme der Beschleunigung gemessen werden
    		// Das Stoppen einer ruckartigen Bewegung, das abrupte Abbremsen wird als negative Beschleunigung
    		// übermittelt und ist betragsmäßig häufig größer als die eigentliche zu beobachtende Beschleunigung.
    		// Die muss bei der Ermittlung dieser Gesten berücksichtigt werden.
    	    if(!_motionUp && !_motionDown) {
                _motionUp = (z >= MOTION_THRESHOLD);
    	    }

    	    if(!_motionUp && !_motionDown) {
                _motionDown = (z <= -MOTION_THRESHOLD);
    	    }
    
    	    if(!_motionLeft && !_motionRight) {
                _motionLeft = (x >= MOTION_THRESHOLD);
    	    }
    	    
    	    if(!_motionRight && !_motionLeft) {
                _motionRight = (x <= -MOTION_THRESHOLD);
    	    }
    
            if(_motionUp && !motionTimer.running) {
                motionTimer.start();
                trace("UP");
                _eventDispatcher.dispatchEvent(new GestureEvent(GestureEvent.JERK_UP));
            }
            
            if(_motionDown && !motionTimer.running) {
                motionTimer.start();
                trace("DOWN");
                _eventDispatcher.dispatchEvent(new GestureEvent(GestureEvent.JERK_DOWN));
            }
            
            if(_motionLeft && !motionTimer.running) {
                motionTimer.start();
                trace("LEFT");
                _eventDispatcher.dispatchEvent(new GestureEvent(GestureEvent.JERK_LEFT));
            }

            if(_motionRight && !motionTimer.running) {
                motionTimer.start();
                trace("RIGHT");
                _eventDispatcher.dispatchEvent(new GestureEvent(GestureEvent.JERK_RIGHT));
            }
        }
        
		//-----------------------------------------------------------------------------------
		// IEventDispatcher
		//-----------------------------------------------------------------------------------
		 
		/**
		 * Registers an event listener object with a GestureDetection object so that the listener receives notification of an event.
		 * 
		 * @param type The type of event.
		 * @param listener The listener function that processes the event.
		 * @param useCapture Determines whether the listener works in the capture phase or the target and bubbling phases.
		 * @param priority The priority level of the event listener.
		 * @param useWeakReference Determines whether the reference to the listener is strong or weak.
		 * 
		 * @see http://livedocs.adobe.com/flex/2/langref/flash/events/IEventDispatcher.html#addEventListener() flash.events.IEventDispatcher.addEventListener()
		 */		
		public function addEventListener(type: String, listener: Function, useCapture: Boolean = false, priority: int = 0, useWeakReference: Boolean = false):void {
			_eventDispatcher.addEventListener(type, listener, useCapture, priority, useWeakReference);
		}

		/**
		 * Dispatches an event into the event flow.
		 * 
		 * @param event The Event object dispatched into the event flow.
		 * 
		 * @see http://livedocs.adobe.com/flex/2/langref/flash/events/IEventDispatcher.html#dispatchEvent() flash.events.IEventDispatcher.dispatchEvent()
		 */
		public function dispatchEvent(event: Event):Boolean {
			return _eventDispatcher.dispatchEvent(event);
		}
		
		/**
		 * Checks whether the GestureDetection object has any listeners registered for a specific type of event.
		 * 
		 * @param type The type of event.
		 * @return A value of <code>true</code> if a listener of the specified type is registered; <code>false</code> otherwise.
		 * 
		 * @see http://livedocs.adobe.com/flex/2/langref/flash/events/IEventDispatcher.html#hasEventListener() flash.events.IEventDispatcher.hasEventListener()
		 */		
		public function hasEventListener(type: String):Boolean {
			return _eventDispatcher.hasEventListener(type);
		}
		
		/**
		 * Removes a listener from the GestureDetection object.
		 * 
		 * @param type The type of event.
		 * @param listener The listener object to remove.
		 * @param useCapture Specifies whether the listener was registered for the capture phase or the target and bubbling phases.
		 * 
		 * @see http://livedocs.adobe.com/flex/2/langref/flash/events/IEventDispatcher.html#removeEventListener() flash.events.IEventDispatcher.removeEventListener()
		 */		
		public function removeEventListener(type: String, listener: Function, useCapture: Boolean = false):void {
			_eventDispatcher.removeEventListener(type, listener, useCapture);
		}
		
		/**
		 * Checks whether an event listener is registered with this GestureDetection object or any of its ancestors for the specified event type.
		 * 
		 * @param type The type of event.
		 * @return A value of <code>true</code> if a listener of the specified type will be triggered; <code>false</code> otherwise.
		 * 
		 * @see http://livedocs.adobe.com/flex/2/langref/flash/events/IEventDispatcher.html#willTrigger() flash.events.IEventDispatcher.willTrigger()
		 */	
		public function willTrigger(type: String):Boolean {
			return _eventDispatcher.willTrigger(type);
		}
        
    }
}